#include "pre.h"
#include "mbranch.h"
#include <malloc.h>
#include <stdio.h>
#include <cstring>
using namespace std;

#define NLAlloc(type,num) (type *)calloc(num,sizeof(type))

bool branch_recog_m;

MBranch::MBranch(int _branch_type, MosList *moshead_, NodeList *nodehead_)
{
	moshead = moshead_;
	nodehead = nodehead_;
	success = 0;
	greatthan4 = 0;
	ready_to_delete = 0;
	num_of_mos = 0;
	branch_type = _branch_type;
	pn_node = NULL;
	headername = NULL;
	mos_list = NULL;
	bak_mos_list = NULL;   // Revised
	node_list = NULL;
	name_list = NULL;
	relation_list = NULL;
	next = NULL;
}

MBranch::~MBranch()
{
	if(moshead) delete moshead;
	if(nodehead) delete nodehead;
	if(mos_list) delete mos_list;
	if(bak_mos_list) delete bak_mos_list;   // Revised
	if(headername) delete headername;
	if(node_list) delete node_list;
	if(name_list) delete name_list;
	if(relation_list) delete relation_list;
	if(next) delete next;
}

char *
MBranch::get_vdd(MosList *moshead)
{
   char *vdd;
   MosList *curr = NULL;
   for(curr=moshead; curr!=NULL; curr=curr->next)
   {
	if(curr->nptype)
	{
	  vdd = NLAlloc(char,strlen(curr->b));
	  strcpy(vdd, curr->b);
	  break;
	}
   }
   return vdd;
}

char *
MBranch::get_gnd(MosList *moshead)
{
   char *gnd;
   MosList *curr = NULL;
   for(curr=moshead; curr!=NULL; curr=curr->next)
   {
        if(!curr->nptype)
        {
          gnd = NLAlloc(char,strlen(curr->b));
          strcpy(gnd, curr->b);
          break;
        }
   }    
   return gnd;
}

int
MBranch::query_moslist(MosList *mos)
{
    int exist_flag = 0;
	
    if(mos_list==NULL)
	return 0;
    else
    {
	for(MosList *cur_m=mos_list; cur_m!=NULL; cur_m=cur_m->next)
	{
	    if(!strcmp(mos->mosname, cur_m->mosname))
	    {
		exist_flag = 1;
		break;
	    }
	}
    }

   if(exist_flag)
        return 1;
   else
        return 0;
}


int
MBranch::node_exist(char *node)
{
   int exist_flag = 0;

   if(node_list==NULL)
	return 0;
   else
   {
	BranchNodeList *curr;
	for(curr=node_list; curr!=NULL; curr=curr->next)
	{
	   if(!strcmp(curr->nodename, node))
		exist_flag = 1; 
	}
   }
   
   if(exist_flag)
	return 1;
   else
	return 0;
}

MosList *
MBranch::insert_mos(MosList *head, MosList *mos)
{

   MosList *current;
   if(head==NULL)
   {
	head = NLAlloc(MosList,1);
	head->mosname = NLAlloc(char,strlen(mos->mosname));
        strcpy(head->mosname,mos->mosname);
        head->d = NLAlloc(char,strlen(mos->d));
        strcpy(head->d,mos->d);
        head->g = NLAlloc(char,strlen(mos->g));
        strcpy(head->g,mos->g);
        head->s = NLAlloc(char,strlen(mos->s));
        strcpy(head->s,mos->s);
        head->nptype = mos->nptype;
	head->is_instead = mos->is_instead;
        head->b = NLAlloc(char,strlen(mos->b));
        strcpy(head->b,mos->b);
        head->modelname = NLAlloc(char,strlen(mos->modelname));
        strcpy(head->modelname,mos->modelname);
        head->lvalue = mos->lvalue;
        head->wvalue = mos->wvalue;
		head->in_branch = mos->in_branch;
	head->next = NULL;
   }
   else
   {
        current = NLAlloc(MosList,1);
        current->mosname = NLAlloc(char, strlen(mos->mosname));
        strcpy(current->mosname,mos->mosname);
        current->d = NLAlloc(char, strlen(mos->d));
        strcpy(current->d,mos->d);
        current->g = NLAlloc(char, strlen(mos->g));
        strcpy(current->g,mos->g);
        current->s = NLAlloc(char, strlen(mos->s));
        strcpy(current->s,mos->s);
        current->nptype = mos->nptype;
	current->is_instead = mos->is_instead;
        current->b = NLAlloc(char, strlen(mos->b));
        strcpy(current->b,mos->b);
        current->modelname = NLAlloc(char, strlen(mos->modelname));
        strcpy(current->modelname,mos->modelname);
        current->lvalue = mos->lvalue;
        current->wvalue = mos->wvalue;
		current->in_branch = mos->in_branch;
        current->next = head->next;
        head->next = current;
   }
   return head;
}	

BranchNodeList *
MBranch::insert_node(BranchNodeList *head, char *node)
{
   BranchNodeList *current;
   if(head==NULL)
   {
	head=NLAlloc(BranchNodeList,1);
	head->nodename = NLAlloc(char, strlen(node));
	strcpy(head->nodename,node);
	head->next = NULL;
   }
   else
   {
	current = NLAlloc(BranchNodeList,1);
	current->nodename = NLAlloc(char, strlen(node));
	strcpy(current->nodename,node);
	current->next = head->next;
	head->next = current;
   }
   return head;
}





BranchNodeList *
MBranch::find_new_node(BranchNodeList *branchnode_list, int p_n_type)
{
   BranchNodeList *new_list = NULL;
   char *vdd = get_vdd(moshead);
   char *gnd = get_gnd(moshead);

   NodeList *curr_node;
   BranchNodeList *curr;

   for(curr=branchnode_list; curr!=NULL; curr=curr->next)
   {
	for(curr_node=nodehead; curr_node!=NULL; curr_node=curr_node->next)
	{
	   if(!strcmp(curr->nodename, curr_node->nodename))
	   {
		for(DeviceList *cur_device=curr_node->devicelist; cur_device!=NULL; cur_device=cur_device->next)
		{
//		   printf("%s  ", cur_device->devicename);
		   char *new_node;
		   int is_pntype = 0;

		   if(!strcmp(curr->nodename,cur_device->d))
		   {
			   new_node = NLAlloc(char, strlen(cur_device->s));
			   strcpy(new_node, cur_device->s);
		   }
		   else if(!strcmp(curr->nodename,cur_device->s))
		   {    
                new_node = NLAlloc(char, strlen(cur_device->d));
                strcpy(new_node, cur_device->d);
           }
		   else if(!strcmp(curr->nodename,cur_device->g))
				new_node = NULL;
		   else
				;

// Revised here, if N branch hit VDD, branch recognition falied. If B branch hit gnd, fail.
		   if(p_n_type && !strcmp(new_node, gnd))
		   {
			  branch_recog_m = false;
			  return NULL;
		   }
		   else if(!p_n_type && !strcmp(new_node, vdd))
		   {
			  branch_recog_m = false;
			  return NULL;
		   }
		   else
				;
// End here

		  if(new_node!=NULL)
		  {
		   for(NodeList *cur_node2=nodehead; cur_node2!=NULL; cur_node2=cur_node2->next)
		   {
			if(!strcmp(cur_node2->nodename, new_node))
			   is_pntype = cur_node2->pn_node;
		   }

		   if( strcmp(new_node,vdd) && strcmp(new_node,gnd) && !node_exist(new_node)) 
		   {
			if(!is_pntype)
			{
			   new_list = insert_node(new_list, new_node);
			   node_list = insert_node(node_list, new_node);
			}
			else
			{
			   pn_node = NLAlloc(char,strlen(new_node));
			   strcpy(pn_node,new_node);
			}
		   }
		  }
		}
	   }
	}
   }
   return new_list;
}
	
bool
MBranch::generate_branch(MosList *mos, int p_n_type)
{
  char *vdd = get_vdd(moshead);
  char *gnd = get_gnd(moshead); 
  char *new_node;

  branch_recog_m = true;

  if(!strcmp(mos->d,vdd) || !strcmp(mos->d,gnd))
  {
	new_node = NLAlloc(char, strlen(mos->s));
        strcpy(new_node, mos->s);
  }
  else
  {
        new_node = NLAlloc(char, strlen(mos->d));
        strcpy(new_node, mos->d);
  }

  int first_is_end = 0;
  for(NodeList *cur_node=nodehead; cur_node!=NULL; cur_node=cur_node->next)
  {
	if(!strcmp(cur_node->nodename, new_node))
	{
	  if(cur_node->pn_node)
		first_is_end = 1;
	  break;
	}
  }

  if(first_is_end)
  {
	mos_list = insert_mos(mos_list, mos);
	pn_node = NLAlloc(char,strlen(new_node));
        strcpy(pn_node,new_node);
//	printf("only one: %s\n", mos->mosname);
  }
  else
  {
	BranchNodeList *new_list = NULL;
	new_list = insert_node(new_list, new_node);
	node_list = insert_node(node_list, new_node);

	while(new_list!=NULL)
	{
	   new_list = find_new_node(new_list, p_n_type);
	   if(!branch_recog_m)
		  return false;
	}
/*
	for(BranchNodeList *cur=node_list; cur!=NULL; cur=cur->next)
	   printf("%s ", cur->nodename);
	printf("\n");
*/

  if(pn_node!=NULL)
  {
	for(BranchNodeList *cur_b=node_list; cur_b!=NULL; cur_b=cur_b->next)
	{
	    for(MosList *cur_mos=moshead; cur_mos!=NULL; cur_mos=cur_mos->next)
		{
    	    if(!strcmp(cur_mos->s,cur_b->nodename) || !strcmp(cur_mos->d,cur_b->nodename))
		    {
				cur_mos->in_branch = true;
//				printf("%s  ", cur_mos->mosname);
				if(!query_moslist(cur_mos))
				    mos_list = insert_mos(mos_list, cur_mos);
		    }
		}
	}
  }
  else   // If there is no pn_node, then the branch recognition fails.
    return false;

/*
	for(MosList *cur_p=mos_list; cur_p!=NULL; cur_p=cur_p->next)
	    printf("%s ", cur_p->mosname);
	printf("\n");
*/
    }	
	return true;
}


NameList *
MBranch::insert_name(NameList *head, char *name, char virtual_name)
{
    NameList *current;
    if(head==NULL)
    {
	head=NLAlloc(NameList,1);
	head->gatename = NLAlloc(char, strlen(name));
	strcpy(head->gatename, name);
	head->virtual_name = virtual_name;
	head->next=NULL;
    }
    else
    {
	current=NLAlloc(NameList,1);
	current->gatename = NLAlloc(char, strlen(name));
	strcpy(current->gatename, name);
	current->virtual_name = virtual_name;
	current->next = head->next;
	head->next = current;
    }
    return head;
}

void 
MBranch::back_up_mos_list()
{
	for(MosList *curr=mos_list; curr!=NULL; curr=curr->next)
	{
		bak_mos_list = insert_mos(bak_mos_list, curr);
	}
}

